# -*- coding: utf-8 -*-
__Author__ = "Saad Ahmad"
__Project__ = "Gravity Code"
__Created__ = "07/31/2018"
__all__ = ['SlimResults']

from scipy import stats
import numpy as np
from statsmodels.compat.python import lzip



class SlimResults(object):
    def __init__(self, glm_results=None):
        '''
        Create a version of the dictionary of results objects that uses less memory. The SlimResults object is a
        smaller subset of the GLMResultsWrapper object in the statsmodels package. Large attributes, such as
        copies of the estimating data, are removed from the results to cut back on memory size.  The results most
        commonly referenced are retained, though.

        The SlimResults object retains only the attributes listed below. For additional information see the
        documentation for the GLMResultsWrapper in the statsmodels package.

        Args:
            glm_results: statsmodels.genmod.generalized_linear_model.GLMResultsWrapper
                An instance of the satasmodels.GLM.fit() results object.

        Attributes:
            params: Pandas Series
                Estimated parameter values
            aic: float
                Akaike Information Criterion
            bic: float
                Bayes Information Criterion
            llf: float
                Value of log-likelihood function
            nobs: float
                number of observations
            bse: Pandas Series
                Beta standard errors for parameter estimates
            pvalues: Pandas Series
                Two-tailed pvalues for parameter estimates
            family_name: str
                Name of distribution family used
            family_link: str
                Estimation link function
            method: str
                Estimation method
            fit_history: int
                Number of iterations completed
            scale: float
                The estimate of the scale / dispersion for the model fit
            deviance: float
                Deviance measure
            pearson_chi2: Pandas Series
                Chi-squared statistic
            cov_type: str
                Covariance type
            yname: str
                Column name of endogenous variable
            xname: List[str]
                Column names of exogenous variables
            model: str
                Model used for fit
            df_resid: float
            df_model: float
            tvalues: Pandas Series
                T statistics
            fittedvalues: Pandas Series
                Linear predicted values

        Methods:
            The SlimResults object replicates two methods from the original GLMResultsWrapper object from statsmodels.

            conf_int: array
                create confidence intervals for parameter estimates.
            summary: object
                print a table summarizing estimation results (replicates statsmodels summary method fo GLM).

        Returns: object
            A SlimResults object.
        '''
        self.params = glm_results.params
        self.aic = glm_results.aic
        self.bic = glm_results.bic
        self.llf = glm_results.llf
        self.nobs = glm_results.nobs
        self.bse = glm_results.bse
        self.pvalues = glm_results.pvalues
        self.family_name=glm_results.family.__class__.__name__
        self.family_link=glm_results.family.link.__class__.__name__
        self.method=glm_results.method
        self.fit_history=glm_results.fit_history['iteration']
        self.scale=glm_results.scale
        self.deviance=glm_results.deviance
        self.pearson_chi2=glm_results.pearson_chi2
        self.cov_type=glm_results.cov_type
        self.yname=glm_results.model.endog_names
        self.xname=glm_results.model.exog_names
        self.model=glm_results.model.__class__.__name__
        self.df_resid=glm_results.df_resid
        self.df_model=glm_results.df_model
        self.tvalues=glm_results.tvalues
        self.fittedvalues = glm_results.fittedvalues
       

    
    def conf_int(self, alpha=.05, cols=None, method='default'):
            """
            Returns the confidence interval of the fitted parameters.

            Parameters
            ----------
            alpha : float, optional
                The significance level for the confidence interval.
                ie., The default `alpha` = .05 returns a 95% confidence interval.
            cols : array-like, optional
                `cols` specifies which confidence intervals to return
            method : string
                Not Implemented Yet
                Method to estimate the confidence_interval.
                "Default" : uses self.bse which is based on inverse Hessian for MLE
                "hjjh" :
                "jac" :
                "boot-bse"
                "boot_quant"
                "profile"


            Returns
            --------
            conf_int : array
                Each row contains [lower, upper] limits of the confidence interval
                for the corresponding parameter. The first column contains all
                lower, the second column contains all upper limits.

            Examples
            --------
            >>> import statsmodels.api as sm
            >>> data = sm.datasets.longley.load()
            >>> data.exog = sm.add_constant(data.exog)
            >>> results = sm.OLS(data.endog, data.exog).fit()
            >>> results.conf_int()
            array([[-5496529.48322745, -1467987.78596704],
                   [    -177.02903529,      207.15277984],
                   [      -0.1115811 ,        0.03994274],
                   [      -3.12506664,       -0.91539297],
                   [      -1.5179487 ,       -0.54850503],
                   [      -0.56251721,        0.460309  ],
                   [     798.7875153 ,     2859.51541392]])


            >>> results.conf_int(cols=(2,3))
            array([[-0.1115811 ,  0.03994274],
                   [-3.12506664, -0.91539297]])

            Notes
            -----
            The confidence interval is based on the standard normal distribution.
            Models wish to use a different distribution should overwrite this
            method.
            """
            bse = self.bse
    
            dist = stats.t
            df_resid = getattr(self, 'df_resid_inference', self.df_resid)
            q = dist.ppf(1 - alpha / 2, df_resid)
    
            if cols is None:
                lower = self.params - q * bse
                upper = self.params + q * bse
            else:
                cols = np.asarray(cols)
                lower = self.params[cols] - q * bse[cols]
                upper = self.params[cols] + q * bse[cols]
            return np.asarray(lzip(lower, upper))
   
    def summary(self, yname=None, xname=None, title=None, alpha=.05):
        """
        Summarize the Regression Results

        Parameters
        -----------
        yname : string, optional
            Default is `y`
        xname : list of strings, optional
            Default is `var_##` for ## in p the number of regressors
        title : string, optional
            Title for the top table. If not None, then this replaces the
            default title
        alpha : float
            significance level for the confidence intervals

        Returns
        -------
        smry : Summary instance
            this holds the summary tables and text, which can be printed or
            converted to various output formats.

        See Also
        --------
        statsmodels.iolib.summary.Summary : class to hold summary
            results

        """

        top_left = [('Dep. Variable:', None),
                    ('Model:', [self.model]),
                    ('Model Family:', [self.family_name]),
                    ('Link Function:', [self.family_link]),
                    ('Method:', [self.method]),
                    ('Covariance Type:', [self.cov_type]),
                    ('No. Observations:', [self.nobs])
                    ]

        top_right = [('No. Iterations:',
                     ["%d" % self.fit_history]),
                     ('Df Residuals:', [self.df_resid]),
                     ('Df Model:', [self.df_model]),
                     ('Scale:', ["%#8.5g" % self.scale]),
                     ('Log-Likelihood:', ["%#8.5g" % self.llf]),
                     ('Deviance:', ["%#8.5g" % self.deviance]),
                     ('Pearson chi2:', ["%#6.3g" % self.pearson_chi2])
                     ]

        if title is None:
            title = "Generalized Linear Model Regression Results"

        # create summary tables
        y=self.yname
        x=self.xname

        from statsmodels.iolib.summary import Summary
        
                    
        smry = Summary()
        smry.add_table_2cols(self, gleft=top_left, gright=top_right,  # [],
                             yname=y, xname=x, title=title)
        smry.add_table_params(self, yname=y, xname=x, alpha=0.05,
                              use_t=True)

        if hasattr(self, 'constraints'):
            smry.add_extra_txt(['Model has been estimated subject to linear '
                                'equality constraints.'])

        # diagnostic table is not used yet:
        # smry.add_table_2cols(self, gleft=diagn_left, gright=diagn_right,
        #                   yname=yname, xname=xname,
        #                   title="")

        return smry






from .combine_sector_results import *
from .DiagnosticsLog import *
from .EstimationModel import *
from .format_regression_table import *
from .save_and_load import *
from .SlimResults import *
from .Specification import *
from .visualize_results import *


__author__ = "Saad Ahmad and Peter Herman"
__project__ = "gme.estimate"
__created__ = "07-31-2018"
__all__ = ['save_estimation', 'load_estimation']

import pickle

def save_estimation(estimation_model, filename: str):
    '''
    Save a serialized copy of an EstimationModel or other object at the specified path.
    Args:
        estimation_model: object
            The object to be saved.
        filename: str
            The path and file name under which to save the object.

    Returns: None

    Examples:
        >>> save_estimation(sample_model, "c:\\Documents\\saved_object.p")

    '''
    f = open(filename, "wb")
    pickle.dump(estimation_model, f)



def load_estimation(filename: str):
    '''
    Load a saved estimation model
    Args:
        filename: str
            The path and file name under which an object was saved.

    Returns: object
        A loaded instance of the specified object.

    Examples:
        >>> loaded_model = load_estimation("c:\\Documents\\saved_object.p")
    '''
    f = open(filename, 'rb')
    estimation_model = pickle.load(f)
    f.close
    return estimation_model



__Author__ = "Peter Herman"
__Project__ = "Gravity Code"
__Created__ = "03/12/2018"
__all__ = ['Specification']

from typing import List

class Specification(object):
    def __init__(self,
                 spec_name:str = 'default_name',
                 lhs_var: str = None,
                 rhs_var: List[str] = None,
                 sector_by_sector: bool = False,
                 drop_imp_exp: List[str] = [],
                 drop_imp: List[str] = [],
                 drop_exp: List[str] = [],
                 keep_imp_exp: List[str] = [],
                 keep_imp: List[str] = [],
                 keep_exp: List[str] = [],
                 drop_years: List[str] = [],
                 keep_years: List[str] = [],
                 drop_missing: bool = False,
                 variables_to_drop_missing: List[str] = [],
                 fixed_effects:List[str] = [],
                 omit_fixed_effect:List[str] = [],
                 std_errors:str = 'HC1',
                 iteration_limit:int = 1000,
                 drop_intratrade:bool = True,
                 cluster: bool=False,
                 cluster_on: str=None,
                 verbose:bool = True):
        if lhs_var is None:
            raise ValueError('lhs_var (left hand side variable) must be specified.')

        self.spec_name = spec_name
        self.lhs_var = lhs_var
        self.rhs_var = rhs_var
        self.sector_by_sector = sector_by_sector
        self.drop_imp_exp = drop_imp_exp
        self.drop_imp = drop_imp
        self.drop_exp = drop_exp
        self.keep_imp_exp = keep_imp_exp
        self.keep_imp = keep_imp
        self.keep_exp = keep_exp
        self.drop_years = drop_years
        self.keep_years = keep_years
        self.drop_missing = drop_missing
        self.variables_to_drop_missing = variables_to_drop_missing
        self.fixed_effects = fixed_effects
        self.omit_fixed_effect = omit_fixed_effect
        self.std_errors = std_errors
        self.iteration_limit = iteration_limit
        self.drop_intratrade = drop_intratrade
        self.cluster=cluster
        self.cluster_on=cluster_on        
        self.verbose = verbose


__Author__ = "Peter Herman"
__Project__ = "Gravity Code"
__Created__ = "03/12/2018"
__all__ = ['EstimationModel']

from typing import List,Tuple, Dict
from typing import Union
from .Specification import Specification
from .DiagnosticsLog import DiagnosticsLog
from .combine_sector_results import combine_sector_results
from ._slice_data_for_estimation import _slice_data_for_estimation
from ._ppml_estimation_and_diagnostics import _estimate_ppml
from .format_regression_table import format_regression_table
from .SlimResults import SlimResults


class EstimationModel(object):
    def __init__(self,
                 estimation_data=None,
                 spec_name: str = 'default_name',
                 lhs_var: str = None,
                 rhs_var: List[str] = [],
                 sector_by_sector: bool = False,
                 drop_imp_exp: List[str] = [],
                 drop_imp: List[str] = [],
                 drop_exp: List[str] = [],
                 keep_imp_exp: List[str] = [],
                 keep_imp: List[str] = [],
                 keep_exp: List[str] = [],
                 drop_years: List[str] = [],
                 keep_years: List[str] = [],
                 drop_missing: bool = True,
                 variables_to_drop_missing: List[str] = None,
                 fixed_effects: List[Union[str, List[str]]] = [],
                 omit_fixed_effect: Dict[Union[str, Tuple[str]],List[Union[str, Tuple[str]]]] = {},
                 std_errors: str = 'HC1',
                 iteration_limit: int = 1000,
                 drop_intratrade: bool = False,
                 retain_modified_data: bool = True,
                 full_results: bool = True,
                 cluster_on: str=None):
        '''
        The GME object is used to specify and run an gravity estimation.  A gme.EstimationData must be supplied along with a
        collection of largely optional arguments that specify variables to include, fixed effects to create, and how
        to perform the regression, among other options.  After the definition of the model, additional methods such as
        .estimate(), which performs the PPML estimation, or .combine_sector_results(), which combines the results for
        each sector (if applicable) can be called.

        Args:
            estimation_data: gme.EstimationData
                A GME EstimationData to use as the basis of the gravity model.
            spec_name: (optional) str
                A name for the model.
            lhs_var: str
                The column name of the variable to be used as the dependent or 'left-hand-side' variable in the
                regression.
            rhs_var: List[str]
                A list of column names for the independent or 'right-hand-side' variable(s) to be used in the regression.
            sector_by_sector: bool
                If true, separate models are estimated for each sector, individually.  Default is False. If True, a
                sector_var_name must have been supplied to the EstimationData.
            drop_imp_exp: (optional) List[str]
                A list of country identifiers to be excluded from the estimation when they appear as an importer or
                exporter
            drop_imp: (optional) List[str]
                A list of country identifiers to be excluded from the estimation when they appear as an importer
            drop_exp: (optional) List[str]
                A list of country identifiers to be excluded from the estimation when they appear as an exporter
            keep_imp_exp: (optional) List[str]
                A list of countries to include in the estimation as either importers or exporters. All others not
                specified are excluded.
            keep_imp: (optional) List[str]
                A list of countries to include in the estimation as importers. All others not specified are excluded.
            keep_exp: (optional) List[str]
                A list of countries to include in the estimation as exporters. All others not specified are excluded.
            drop_years: (optional) list
                A list of years to exclude from the estimation. The list elements should match the dtype of the year
                column in the EstimationData.
            keep_years: (optional) list
                A list of years to include in the estimation. The list elements should match the dtype of the year
                column in the EstimationData.
            drop_missing: bool
                If True, rows with missing values are dropped. Default is true, which drops if observations are missing
                in any of the columns specified by lhs_var or rhs_var.
            variables_to_drop_missing: (optional) List[str]
                A list of column names for specifying which columns to check for missing values when dropping rows.
            fixed_effects: (optional) List[Union[str,List[str]]]
                A list of variables to construct fixed effects based on. Can accept single string entries, which create
                fixed effects corresponding to that variable or lists of strings that create fixed effects corresponding
                to the interaction of the list items. For example, fixed_effects = ['importer',['exporter','year']]
                would create a set of importer fixed effects and a set of exporter-year fixed effects.
            omit_fixed_effect: (optional) Dict[Union[str,Tuple[str]]:List[str, Tuple[str]]]
                A dictionary of fixed effect categories and values to be dropped from dataframe. The dictionary key(s)
                can be either a single string corresponding to one dimension of the fixed effects or a tuple of strings
                specifying multiple dimesions. The values should then be either a list of strings for 1 dimension or a
                list of tuples for multiple dimensions. For example: {'importer':['DEU', 'ZAF'], 'exporter':['USA']} or
                {('importer','year'):[('ARG','2015'),('BEL','2013')]}
                The fixed effect categories in the keys need to be a subset of the list supplied for fixed_effects.
                If not specified, the colinearity diagnostics will identify a column to drop on its own.
            std_errors: (optional) str
                Specifies the type of standard errors to be computed. Default is HC1, heteroskedascticity robust errors.
                See http://www.statsmodels.org/dev/generated/statsmodels.regression.linear_model.RegressionResults.html
                for alternative options.
            iteration_limit: (optional) int
                Upper limit on the number of iterations for the estimation procedure. Default is 1000.
            drop_intratrade: (optional) bool
                If True, intra-national trade flows (importer == exporter) are excluded from the regression. Default is
                False.
            retain_modified_data: (optional) bool
                If True, the estimation DataFrames for each sector after they have been (potentially) modified during
                the pre-diagnostics for collinearity and convergence issues. Default is False. WARNING: these object
                sizes can be very large in memory so use with caution.
            full_results: bool
                If True, estimate() returns the full results object from the GLM estimation.  These results can be quite
                large as each estimated sector's results will contain a full copy of the data used for its estimation,
                vectors of predicted values, and other memory intensive pieces of data.  If False, estimate() returns
                a smaller subset of the results that are likely most useful (e.g. .params, .nobs, .bse, .pvalues, .aic,
                .bic).  For a list of these attributes, see the documentation for the function SlimResults
            cluster_on: (optional) str
                The name of a column of categorical variables to use as clusters for clustered standard errors.

        Attributes:
            estimation_data: Return the EstimationData.
            results_dict: Return the dictionary of regression results (after applying estimate method).
            modified_data: Return data modified data after removing problematic columns (after applying estimate method)
            ppml_diagnostics: Return PPML estimation diagnostic information (after applying estimate method)


        Methods:
            estimate: Estimate a PPML model.
            combine_sector_results: Combine multiple result_dict entries into a single DataFrame.
            format_regression_table: Format regression results into a text, csv, or LaTeX table for presentation.


        Examples:
            Declare an EstimationModel
             >>> sample_estimation_model = EstimationModel(estimation_data = gme_data,
             ...                                           lhs_var = 'trade_value',
             ...                                           rhs_var = ['log_distance',
             ...                                           'agree_pta',
             ...                                           'common_language',
             ...                                           'contiguity'])
            Estimate the model
            >>> sample_estimation_model.estimate("ppml")

            Extract the results
            >>> results_dictionary = sample_estimation_model.results_dict

            Write the estimates, p-values, and std. errors from all sectors to a .csv file.
            >>> sample_estimation_model.combine_sector_results("c:\\folder\\saved_results.csv")

            Create and export a formatted table of estimation results
            >>> sample_estimation_model.format_regression_table(format = 'csv',
            ...                                                 path = "c:\\folder\\saved_results.csv")

            Define a model with fixed effects and drop specific ones
            >>> fe_model = EstimationModel(est_data,
                                           lhs_var='trade_value',
                                           rhs_var=grav_vars,
                                           fixed_effects=[['importer', 'year'], ['exporter', 'year']],
                                           omit_fixed_effect={('importer','year'):[('DEU','2015')],'exporter':['USA']})

        '''

        # Value checks for inputs
        if estimation_data is None:
            raise ValueError("A EstimationData must be provided.")

        if cluster_on is not None and not isinstance(cluster_on, str):
            raise ValueError("cluster_on must be a string indicating the column to cluster on.")

        if not isinstance(omit_fixed_effect, dict):
            raise ValueError("As of gme v1.3, omit_fixed_effect argument must be a dictionary.")

        ##############
        # Attributes #
        ##############
        if (variables_to_drop_missing is None) & (drop_missing):
            variables_to_drop_missing = [lhs_var] + rhs_var
        self.estimation_data = estimation_data
        self.data_log = estimation_data.data_log

        # Create indicator for clustering
        if cluster_on is not None:
            cluster = True
        else:
            cluster = False

        self.specification = Specification(spec_name=spec_name,
                                           lhs_var=lhs_var,
                                           rhs_var=rhs_var,
                                           sector_by_sector=sector_by_sector,
                                           drop_imp_exp=drop_imp_exp,
                                           drop_imp=drop_imp,
                                           drop_exp=drop_exp,
                                           keep_imp_exp=keep_imp_exp,
                                           keep_imp=keep_imp,
                                           keep_exp=keep_exp,
                                           drop_years=drop_years,
                                           keep_years=keep_years,
                                           drop_missing=drop_missing,
                                           variables_to_drop_missing=variables_to_drop_missing,
                                           fixed_effects=fixed_effects,
                                           omit_fixed_effect=omit_fixed_effect,
                                           std_errors=std_errors,
                                           iteration_limit=iteration_limit,
                                           drop_intratrade=drop_intratrade,
                                           cluster=cluster,
                                           cluster_on=cluster_on,
                                           verbose=False)
        self.retain_modified_data = retain_modified_data
        self.full_results = full_results
        self.results_dict = None
        self.modified_data = None
        self.diagnostics_log = DiagnosticsLog(spec_name=spec_name)
        self.ppml_diagnostics = None

    ###########
    # Methods #
    ###########

    def estimate(self):
        '''
        Perform sector by sector GLM estimation with PPML diagnostics. The routine follows several steps.
        services. If sector_by_sector is specified, the routine is repeated for each sector.

        1. Create Fixed effects: Fixed effects are created based on the EstimationModel specification.

        2. Pre-Diagnostics: Several steps are taken to increase the likelihood that the estimation will converge
            successfully.
            a. Perfect Colinearity: Columns and observations that are perfectly collinear are identified and
                excluded.
            b. Insufficient Variation: Fixed effects categories in which there is an insufficient level of
                variation for estimation are excluded.

        3. Estimate: A GLM, PPML estimation is run.

        4. Post-Diagnostics: A test for over-fit values is conducted.

        The method returns one object and stores three in the EstimationModel.  The first object, which is
        both returned and stored as EstimationModel.results_dict is

        Returns: Dict[statsmodels.genmod.generalized_linear_model.GLMResultsWrapper]
            The primary return is a dictionary of results objects from the statsmodels GLM.fit routine, each keyed using
            either the name of the sector if the estimation was sector by sector (i.e. sector_by_sector = True) or with
            the key 'all' if not.

            Additionally, the method assigns two or three attributes of the Estimation model:
                1. EstimationModel.results_dict:  Dict[statsmodels.genmod.generalized_linear_model.GLMResultsWrapper]
                    A copy of the returned dictionary of results, as described above, is stored with the EstimationModel
                2. EstimationModel.ppml_diagnostics: pandas.core.frame.DataFrame
                    A data frame containing a column of pre- and post-diagnostic information for each regression
                3. EstimationModel.modified_data: (optional) Dict[pandas.core.frame.DataFrame]
                    A dictionary using the same keys as results_dict, each containing the modified dataframes created
                    during the pre-diagnostic stages of the estimations. Because of the large memory footprint of this
                    assignment, it is only done if specified (i.e. EstimationModel.retain_modified_data = True)

        Examples:
            >>> sample_estimation_model.estimate("ppml")
        '''
        data_object = self.estimation_data
        specification = self.specification
        data_log = data_object.data_log
        data_frame = data_object.data_frame
        meta_data = data_object.meta_data
        diagnostics_log = self.diagnostics_log

        if specification.sector_by_sector is True and meta_data.sector_var_name is None:
            raise ValueError('sector_var_name must be specified for sector_by_sector option')

        estimation_data_frame, data_log = _slice_data_for_estimation(data_frame=data_frame,
                                                                      specification=specification,
                                                                      meta_data=meta_data,
                                                                      data_log=data_log)

        results_dict, ppml_diagnostics, modified_data \
            = _estimate_ppml(data_frame=estimation_data_frame,
                             meta_data=meta_data,
                             specification=specification,
                             fixed_effects=specification.fixed_effects,
                             drop_fixed_effect=specification.omit_fixed_effect,
                             cluster=specification.cluster,
                             cluster_on=specification.cluster_on)

        self.ppml_diagnostics = ppml_diagnostics
        if self.retain_modified_data:
            self.modified_data = modified_data

        if self.full_results:
            self.results_dict = results_dict
        else:
            slim_results_dict = {}
            for key in results_dict:
                slim_results_dict[key] = SlimResults(results_dict[key])
            self.results_dict = slim_results_dict


        

        return self.results_dict

    def combine_sector_results(self,
                               write_path: str = None):
        '''
        Combines all results in the results_dict into a single data_frame, retaining estimated coefficients, standard
        errors, and p-values.

        See combine_sector_results() function for more details.

        Examples:
            Create a new DataFrame with results.
            >>> combined_results = sample_estimation_model.combine_sector_results()

            Write to combined results to a .csv file.
            >>> sample_estimation_model.combine_sector_results("c:\folder\saved_results.csv")

            Or do both.
            >>> combined_results = sample_estimation_model.combine_sector_results("c:\folder\saved_results.csv")

        '''
        if self.results_dict is None:
            raise ValueError('EstimationModel must be estimated first (EstimationModel.results_dict cannot be None)')
        return combine_sector_results(self.results_dict, write_path)

    # Commented out until it can be debugged
    # def pickle_model(self,
    #               write_path: str = None):
    #    '''
    #    '''
    #    Save an EstimationModel as a python-readable object (pickle) at a specified location.
    #    Args:
    #        write_path: (str)
    #            The location (including file name) to write the EstimationModel object
    #
    #    Examples:
    #        Save the EstimationModel.
    #        >>> sample_estimation_model.pickle_model("c:\folder\saved_object.p")
    #        Reload the model.
    #        >>> import pickle
    #        >>> loaded_model = pickle.load( open("c:\folder\saved_object.p"))
    #    '''
    #    '''
    #    if write_path is None:
    #        raise ValueError("write_path must be specified.")
    #
    #    pickle.dump(self, open(write_path, "wb"))
    #    '''

    def format_regression_table(self,
                                variable_list: List[str] = [],
                                format: str = 'txt',
                                se_below: bool = True,
                                significance_levels: List[float] = [0.01, 0.05, 0.10],
                                round_values: int = 3,
                                omit_fe_prefix: List[str] = [],
                                table_columns: list = [],
                                path: str = None,
                                include_index: bool = False,
                                latex_syntax: bool = False,
                                r_squared: bool = False,
                                note:str = None
                                ):
        '''
        Format estimation results into a table with significance stars, rounded values, etc. The EstimationModel method
        is a shortcut to the stand-alone function. See the function gme.estimate.format_regression_table for more
        details.

        See format_regression_table() function definition for details.

        Examples:
            Create a .csv file.
            >>> sample_estimation_model.format_regression_table(format = 'csv',
                                                                path = "c:\folder\saved_results.csv")

            Create a LaTeX .tex table without fixed effects (with prefix 'imp_fe_' and 'exp_fe_')
            >>> sample_estimation_model.format_regression_table(format = 'tex',
                                                                path = "c:\folder\saved_results.tex",
                                                                omit_fe_prefix = ['imp_fe_' , 'exp_fe_'])
        '''
        return format_regression_table(results_dict=self.results_dict,
                                       variable_list=variable_list,
                                       format=format,
                                       se_below=se_below,
                                       significance_levels=significance_levels,
                                       round_values=round_values,
                                       omit_fe_prefix=omit_fe_prefix,
                                       table_columns=table_columns,
                                       path=path,
                                       include_index=include_index,
                                       latex_syntax=latex_syntax,
                                       r_squared=r_squared,
                                       note = note)






__Author__ = "Peter Herman"
__Project__ = "Gravity Code"
__Created__ = "03/28/2018"



def _slice_data_for_estimation(data_frame,
                               specification,
                               meta_data,
                               data_log):

    # Keep Only Using Variables
    data_frame, data_log = _keep_using_vars(data_frame, specification, meta_data, data_log)

    # Drop intra-trade
    data_frame, data_log =_drop_intra_trade(data_frame, specification, meta_data, data_log)

    # Drop Importers
    data_frame, data_log = _drop_importers(data_frame, specification, meta_data, data_log)

    # Drop Exporters
    data_frame, data_log = _drop_exporters(data_frame, specification, meta_data, data_log)

    # Keep importers
    data_frame, data_log = _keep_importers(data_frame, specification, meta_data, data_log)

    # Keep Exporters
    data_frame, data_log = _keep_exporters(data_frame, specification, meta_data, data_log)

    # Drop Years
    data_frame, data_log = _drop_years(data_frame, specification, meta_data, data_log)

    # Keep Years
    data_frame, data_log = _keep_years(data_frame, specification, meta_data, data_log)

    # Drop Missing
    data_frame, data_log = _drop_missing(data_frame, specification, data_log)

    return data_frame, data_log



#---------------
# Sub-routines
#---------------

def _keep_using_vars(data_frame,
                               specification,
                               meta_data,
                               data_log):
    pre_drop_size = data_frame.shape

    using_variables = specification.rhs_var + [specification.lhs_var, meta_data.imp_var_name,
                                             meta_data.exp_var_name, meta_data.year_var_name]
    if specification.cluster_on is not None:
        using_variables.append(specification.cluster_on)
    if meta_data.sector_var_name is not None:
        using_variables = using_variables + [meta_data.sector_var_name]
    data_frame = data_frame[using_variables]
    post_drop_size = data_frame.shape
    dropped_dims = {'rows': pre_drop_size[0] - post_drop_size[0], 'columns': pre_drop_size[1] - post_drop_size[1]}
    data_log.specification_variables_kept = str(using_variables) + ', Observations excluded by user: ' + str(dropped_dims)
    if specification.verbose is True:
        print('select specification variables: ' + data_log.specification_variables_kept)
    return data_frame, data_log

def _drop_intra_trade(data_frame,
                     specification,
                     meta_data,
                     data_log):
    pre_drop_size = data_frame.shape
    if specification.drop_intratrade is True:
        data_frame = data_frame[data_frame[meta_data.imp_var_name] != data_frame[meta_data.exp_var_name]].copy()
        data_log.intra_country_trade_dropped = 'yes'
    else:
        data_log.intra_country_trade_dropped = 'no'
    post_drop_size = data_frame.shape
    dropped_dims = {'rows': pre_drop_size[0] - post_drop_size[0], 'columns': pre_drop_size[1] - post_drop_size[1]}
    data_log.intra_country_trade_dropped = data_log.intra_country_trade_dropped + ', Observations excluded by user: ' + str(dropped_dims)
    if specification.verbose is True:
        print('drop_intratrade: ' + data_log.intra_country_trade_dropped)
    return data_frame, data_log


def _drop_importers(data_frame,
                     specification,
                     meta_data,
                     data_log):
    pre_drop_size = data_frame.shape
    if (len(specification.drop_imp_exp) > 0) or (len(specification.drop_imp) > 0):
        importer_drop_list = list(set(specification.drop_imp_exp + specification.drop_imp))
        data_frame.drop(data_frame.loc[data_frame[meta_data.imp_var_name].isin(importer_drop_list)].index, inplace=True)
        data_log.importers_dropped = str(importer_drop_list)
    else:
        data_log.importers_dropped = 'none'
    post_drop_size = data_frame.shape
    dropped_dims = {'rows': pre_drop_size[0] - post_drop_size[0], 'columns': pre_drop_size[1] - post_drop_size[1]}
    data_log.importers_dropped = data_log.importers_dropped + ', Observations excluded by user: ' + str(dropped_dims)
    if specification.verbose is True:
        print('drop_imp: ' + data_log.importers_dropped)
    return data_frame, data_log


def _drop_exporters(data_frame,
                     specification,
                     meta_data,
                     data_log):
    pre_drop_size = data_frame.shape
    if (len(specification.drop_imp_exp) > 0) or (len(specification.drop_exp) > 0):
        exporter_drop_list = list(set(specification.drop_imp_exp + specification.drop_exp))
        data_frame.drop(data_frame.loc[data_frame[meta_data.exp_var_name].isin(exporter_drop_list)].index, inplace=True)
        data_log.exporters_dropped = str(exporter_drop_list)
    else:
        data_log.exporters_dropped = 'none'
    post_drop_size = data_frame.shape
    dropped_dims = {'rows': pre_drop_size[0] - post_drop_size[0], 'columns': pre_drop_size[1] - post_drop_size[1]}
    data_log.exporters_dropped = data_log.exporters_dropped + ', Observations excluded by user: ' + str(dropped_dims)
    if specification.verbose is True:
        print('drop_exp: ' + data_log.exporters_dropped)
    return data_frame, data_log

def _keep_importers(data_frame,
                     specification,
                     meta_data,
                     data_log):
    pre_drop_size = data_frame.shape
    if (len(specification.keep_imp_exp) > 0) or (len(specification.keep_imp) > 0):
        importer_keep_list = list(set(specification.keep_imp_exp + specification.keep_imp))
        data_frame = data_frame.loc[data_frame[meta_data.imp_var_name].isin(importer_keep_list)]
        data_log.importers_kept = str(importer_keep_list)
    else:
        data_log.importers_kept = 'all available'
    post_drop_size = data_frame.shape
    dropped_dims = {'rows': pre_drop_size[0] - post_drop_size[0], 'columns': pre_drop_size[1] - post_drop_size[1]}
    data_log.importers_kept = data_log.importers_kept + ', Observations excluded by user: ' + str(dropped_dims)
    if specification.verbose is True:
        print('keep_imp: ' + data_log.importers_kept)
    return data_frame, data_log

def _keep_exporters(data_frame,
                     specification,
                     meta_data,
                     data_log):
    pre_drop_size = data_frame.shape
    if (len(specification.keep_imp_exp) > 0) or (len(specification.keep_exp) > 0):
        exporter_keep_list = list(set(specification.keep_imp_exp + specification.keep_exp))
        data_frame = data_frame.loc[data_frame[meta_data.exp_var_name].isin(exporter_keep_list)]
        data_log.exporters_kept = str(exporter_keep_list)
    else:
        data_log.exporters_kept = 'all available'
    post_drop_size = data_frame.shape
    dropped_dims = {'rows': pre_drop_size[0] - post_drop_size[0], 'columns': pre_drop_size[1] - post_drop_size[1]}
    data_log.exporters_kept = data_log.exporters_kept + ', Observations excluded by user: ' + str(dropped_dims)
    if specification.verbose is True:
        print('keep_exp: ' + data_log.exporters_kept)
    return data_frame, data_log

def _drop_years(data_frame,
                     specification,
                     meta_data,
                     data_log):
    pre_drop_size = data_frame.shape
    if len(specification.drop_years) > 0:
        data_frame = data_frame.loc[data_frame[meta_data.year_var_name].isin(specification.drop_years)]
        data_log.years_dropped = str(specification.dropped_years)
    else:
        data_log.years_dropped = 'none'
    post_drop_size = data_frame.shape
    dropped_dims = {'rows': pre_drop_size[0] - post_drop_size[0], 'columns': pre_drop_size[1] - post_drop_size[1]}
    data_log.years_dropped = data_log.years_dropped + ', Observations excluded by user: ' + str(dropped_dims)
    if specification.verbose is True:
        print('drop_years: ' + data_log.years_dropped)
    return data_frame, data_log

def _keep_years(data_frame,
                     specification,
                     meta_data,
                     data_log):
    pre_drop_size = data_frame.shape
    if len(specification.keep_years) > 0:
        data_frame = data_frame.loc[data_frame[meta_data.year_var_name].isin(specification.keep_years)]
        data_log.years_kept = str(specification.keep_years)
    else:
        data_log.years_kept = 'all available'
    post_drop_size = data_frame.shape
    dropped_dims = {'rows': pre_drop_size[0] - post_drop_size[0], 'columns': pre_drop_size[1] - post_drop_size[1]}
    data_log.years_kept = data_log.years_kept + ', Observations excluded by user: ' + str(dropped_dims)
    if specification.verbose is True:
        print('keep_years: ' + data_log.years_kept)
    return data_frame, data_log

def _drop_missing(data_frame,
                     specification,
                     data_log):
    pre_drop_size = data_frame.shape
    if specification.drop_missing is True:
        data_frame = data_frame.loc[data_frame.notnull().all(axis = 1)]
        data_log.missing_dropped = 'yes'
    else:
        data_log.missing_dropped = 'no'
    post_drop_size = data_frame.shape
    dropped_dims = {'rows': pre_drop_size[0] - post_drop_size[0], 'columns': pre_drop_size[1] - post_drop_size[1]}
    data_log.missing_dropped = data_log.missing_dropped + ', Observations excluded by user: ' + str(dropped_dims)
    if specification.verbose is True:
        print('drop_missing: ' + data_log.missing_dropped)
    return data_frame, data_log



__Author__ = "Peter Herman"
__Project__ = "GE Gravity"
__Created__ = "02/20/2018"
__all__ = ['EstimationData']
import numpy as np
import pandas as pd
from .DataLog import DataLog
from typing import List
import matplotlib.pyplot as plt


class EstimationData(object):
    def __init__(self,
                 data_frame=None,
                 name: str = 'unnamed',
                 imp_var_name: str = 'importer',
                 exp_var_name: str = 'exporter',
                 year_var_name: str = 'year',
                 trade_var_name: str = None,
                 sector_var_name: str = None,
                 expend_var_name: str = None,
                 output_var_name: str = None,
                 notes: List[str] = list()):
        '''
        An object used for storing data for gravity modeling and producing some summary statistics.
        Args:
            data_frame: Pandas.DataFrame
                A DataFrame containing trade, gravity, etc. data.
            name: (optional) str
                A name for the dataset.
            imp_var_name: str
                The name of the column containing importer IDs.
            exp_var_name: str
                The name of the column containing exporter IDs.
            year_var_name: str
                The name of the column containing year data.
            trade_var_name: (optional) str
                The name of the column containing trade data.
            sector_var_name: (optional) str
                The name of the column containing sector/industry/product IDs.
            expend_var_name: (optional) str
                The name of the column containing importer total expenditure information.
            output_var_name: (optional) str
                The name of the column containing exporter total output information
            notes: (optional) str
                A string to be included as a note n the object.

        Attributes:
            data_frame: Pandas.DataFrame
                The supplied DataFrame.
            name: str
                The supplied data name.
            imp_var_name: str
                The name of the column containing importer IDs.
            exp_var_name: str
                The name of the column containing exporter IDs.
            year_var_name: str
                The name of the column containing year data.
            trade_var_name: str
                The name of the column containing trade data.
            sector_var_name: str
                The name of the column containing sector/industry/product IDs.
            notes: List[str]
                A list of notes.
            number_of_exporters: int
                The number of unique exporter IDs in the dataset.
            number_of_importers: int
                The number of unique importer IDs in the dataset.
            shape: List[int]
                The dimensions of the dataset.
            countries: List[str]
                A list of the unique country IDs in the dataset.
            number_of_countries: int
                The number of unique country IDs in the dataset.
            number_of_years: int
                The number of years in the dataset
            columns: List[str]
                A list of column names in the dataset.
            number_of_sectors: int
                If a sector is specified, the number of unique sector IDs in the dataset.

        Methods:
            tablulate_by_group:
                Summarize columns by a user-specified grouping. Can be used to tabulate, aggregate, summarize,
                etc. data.

                Args:
                    tab_variables: List[str]
                        Column names of variables to be tabulated
                    by_group: List[str]
                        Column names of variables by which to group observations for tabulation.
                    how: List[str]
                        The method by which to combine observations within a group. Can accept 'count','mean',
                        'median','min','max','sum','prod','std', and 'var'. It may work with other numpy or
                        pandas functions.

                Returns: Pandas.DataFrame
                    A DataFrame of tabulated values for each group.

            year_list:
                Returns a list of years present in the data
                Args:
                    none
                Returns: list

            countries_each_year
                Returns a dictionary keyed by year ID containing a list of country IDs present in each corresponding
                year.
                Args:
                    none
                Returns: dict

            sector_list
                Returns a list of unique sector IDs

            dtypes:
                Returns the data types of the columns in the DataObject.data_frame using Pandas.DataFrame.dtypes().
                See Pandas documentation for more information.
                Examples:

            info:
                Print summary information about DataObject.data_frame using Pandas.DataFrame.dtypes(). See Pandas
                documentation for more information.

            describe:
                Generates some descriptive statistics for DataObject.data_frame using Pandas.DataFrame.describe().
                See Pandas documentation for more information.

            add_note:
                Add a note to the list of notes in 'notes' attribute.
                Args:
                     note: str
                        A note to add to DataObject.
                Returns: none

            correlation:
                Return and plot the correlation matrix of the data.
                Args:
                    columns: List[str] (optional)
                        A list of column names to include in the matrix. Default is to include all columns.
                    plot: bool
                        If True, it plots a heatmap of the correlation matrix.
                Returns: Pandas.DataFrame
                    A correlation matrix

            add_pair_var:
                Create a new variable that is a concatenation of categorical variables for use as a fixed effect or
                cluster category, for example. Original values are separated by '_' in the concatenated value.
                Args:
                    var_list: List[str]
                        List of names of columns to concatinate.
                    var_name: (Optional) str
                        Column name to use for the new variable. By default, it uses the names in var_list separated by
                        '_'.
                    symmetric: bool
                        If False, all variables are concatenated in the order of the var_list (e.g. ARG_USA
                        and USA_ARG would both would be created). If True, ignores ordering of values across columns and
                        concatenates alphabetically (e.g. all ARG and USA pairings would concatenate as ARG_USA)

                Returns: None, adds new column to EstimationData.dataframe.


        Examples:
            # Load a DataFrame
            >>> import pandas as pd
            >>> gravity_data = pd.read_csv('https://www.usitc.gov/data/gravity/example_trade_and_grav_data_small.csv')
            >>> gravity_data.head(5)
              importer exporter  year   trade_value  agree_pta  common_language  \
            0      AUS      BRA  1989  3.035469e+08        0.0              1.0
            1      AUS      CAN  1989  8.769946e+08        0.0              1.0
            2      AUS      CHE  1989  4.005245e+08        0.0              1.0
            3      AUS      DEU  1989  2.468977e+09        0.0              0.0
            4      AUS      DNK  1989  1.763072e+08        0.0              1.0
               contiguity  log_distance
            0         0.0      9.553332
            1         0.0      9.637676
            2         0.0      9.687557
            3         0.0      9.675007
            4         0.0      9.657311

            # Create a DataObject
            example_data_object = DataObject(data_frame=gravity_data,
                                             imp_var_name='importer',
                                             exp_var_name='exporter',
                                             trade_var_name='trade_value',
                                             year_var_name='year',
                                             notes='Downloaded from https://www.usitc.gov/data/gravity/example_trade_and_grav_data_small.csv')

            # tabulate_by_group
            # Sum trade value by importer and year
            >>> aggregated_data = example_data_object.tabulate_by_group(tab_variables = ['trade_value'],
            ...                                                         by_group = ['importer', 'year'],
            ...                                                         how = ['sum'])
            >>> aggregated_data.head(5)
              importer_  year_  trade_value_sum
            0       ARG   1989     0.000000e+00
            1       ARG   1990     0.000000e+00
            2       ARG   1991     0.000000e+00
            3       ARG   1992     0.000000e+00
            4       ARG   1993     1.593530e+10

            # Summarize minimum and maximum trade flows between each trading pair
            >>> summarized_data = example_data_object.tabulate_by_group(tab_variables = ['trade_value'],
            ...                                                         by_group = ['importer', 'exporter'],
            ...                                                         how = ['min','max'])
            >>> summarized_data.head(5)
              importer_ exporter_  trade_value_min  trade_value_max
            0       ARG       AUS              0.0     4.095529e+08
            1       ARG       AUT              0.0     2.986187e+08
            2       ARG       BEL              0.0     7.669537e+08
            3       ARG       BOL              0.0     2.743706e+09
            4       ARG       BRA              0.0     2.218091e+10

            # year_list
            >>> example_data_object.year_list()
            [1989,
             1990,
             1991,
             1992,
             ...

            # countries_each_year
            >>> countries = example_data_object.countries_each_year()
            >>> countries.keys()
            dict_keys([1989, 1990, 1991, 1992, 1993, 1994, 1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015])
            >>> countries[1989]
            ['ESP',
             'SGP',
             'PHL',
             'NGA',
             'VEN',
             ...

            # dtypes
            >>> example_data_object.dtypes()
            importer            object
            exporter            object
            year                 int64
            trade_value        float64
            agree_pta          float64
            common_language    float64
            contiguity         float64
            log_distance       float64
            dtype: object

            >>> example_data_object.info()
            <class 'pandas.core.frame.DataFrame'>
            RangeIndex: 98612 entries, 0 to 98611
            Data columns (total 8 columns):
            importer           98612 non-null object
            exporter           98612 non-null object
            year               98612 non-null int64
            trade_value        98612 non-null float64
            agree_pta          97676 non-null float64
            common_language    97676 non-null float64
            contiguity         97676 non-null float64
            log_distance       97676 non-null float64
            dtypes: float64(5), int64(1), object(2)
            memory usage: 6.0+ MB

            # describe
            >>> example_data_object.describe()
                           year   trade_value     agree_pta  common_language  \
            count  98612.000000  9.861200e+04  97676.000000     97676.000000
            mean    2002.210441  1.856316e+09      0.381547         0.380646
            std        7.713050  1.004735e+10      0.485769         0.485548
            min     1989.000000  0.000000e+00      0.000000         0.000000
            25%     1996.000000  1.084703e+06      0.000000         0.000000
            50%     2002.000000  6.597395e+07      0.000000         0.000000
            75%     2009.000000  6.125036e+08      1.000000         1.000000
            max     2015.000000  4.977686e+11      1.000000         1.000000

                     contiguity  log_distance
            count  97676.000000  97676.000000
            mean       0.034051      8.722631
            std        0.181362      0.818818
            min        0.000000      5.061335
            25%        0.000000      8.222970
            50%        0.000000      9.012502
            75%        0.000000      9.303026
            max        1.000000      9.890765

            # notes
            >>> example_data_object.add_note('year IDs are integers')
            >>> example_data_object.notes
            ['Downloaded from https://www.usitc.gov/data/gravity/example_trade_and_grav_data_small.csv',
            'year IDs are integers']


        '''

        ##############
        # Attributes #
        ##############
        self.data_frame = data_frame.copy()
        self.name = name
        self.imp_var_name = imp_var_name
        self.exp_var_name = exp_var_name
        self.year_var_name = year_var_name
        self.trade_var_name = trade_var_name
        self.sector_var_name = sector_var_name
        self.expend_var_name = expend_var_name
        self.output_var_name = output_var_name
        self.notes = [notes]
        self.number_of_exporters = len(data_frame[exp_var_name].unique())
        self.number_of_importers = len(data_frame[imp_var_name].unique())
        self.shape = self.data_frame.shape

        countries = np.concatenate((data_frame[imp_var_name].unique(), data_frame[exp_var_name].unique()),
                                   axis=0)
        countries = pd.DataFrame(pd.DataFrame(countries)[0].unique())
        self.countries = list(countries)
        self.number_of_countries = len(countries)

        self.number_of_years = len(data_frame[year_var_name].unique())
        self.columns = list(self.data_frame.columns.values)
        if sector_var_name is not None:
            self.number_of_sectors = len(data_frame[sector_var_name].unique())
        else:
            self.number_of_sectors = 'not_applicable'

        self.data_log = DataLog()  # Add initialization values as needed

        self.meta_data = _MetaData(imp_var_name=self.imp_var_name,
                                   exp_var_name=self.exp_var_name,
                                   year_var_name=self.year_var_name,
                                   trade_var_name=self.trade_var_name,
                                   sector_var_name=self.sector_var_name,
                                   expend_var_name=self.expend_var_name,
                                   output_var_name=self.output_var_name)

    ###########
    # Methods #
    ###########
    def tabulate_by_group(self, tab_variables: list, by_group: list, how: list):
        '''
        Creates summary stats by group.
        Args:
            tab_variables: List[str]
                A list of column names to tabulate/aggregate
            by_group: List[str]
                A list of column names by which to group the tabulation/aggregation
            how: List[str]
                A list of methods to apply to each tabulated variable. Can be selected from 'count', 'mean','median',
                'min','max','sum','prod','std',or 'var'. It may work with other numpy or pandas functions.
        Returns: Pandas DataFrame
            Tabulated/aggregated results,
        '''
        tabulate_data = self.data_frame[tab_variables + by_group]
        # IDEA: it would be possible to do a version in which a lambda function is fed that can create custom summary stats
        if by_group != []:
            tabulate_data = tabulate_data.groupby(by_group).agg(how).reset_index()
            # The following combines the auto-generated, hierarchic column names into a single name
            tabulate_data.columns = ['_'.join(col).strip() for col in tabulate_data.columns.values]
        else:
            tabulate_data = tabulate_data.agg(how).reset_index()
        return tabulate_data

    def dtypes(self):
        return self.data_frame.dtypes

    def info(self):
        return self.data_frame.info()

    def describe(self):
        return self.data_frame.describe()

    def year_list(self):
        '''
        A function that returns a list of the years present in the dataframe
        Returns: list
            A list of years
        '''
        return list(self.data_frame[self.year_var_name].unique())

    def countries_each_year(self):
        '''
        A function to create a collection of countries present in each year of the data.
        Returns: dict
            A dictionary in which each dictionary entry corresponds to a year and features a (list) of countries.
            i.e. {year:country_list}. For example: return_object[year] returns a list of countries in that year
        '''
        country_dict = {}
        year_list = list(self.data_frame[self.year_var_name].unique())
        for year in year_list:
            year_data = self.data_frame.loc[self.data_frame[self.year_var_name] == year]
            importers = list(year_data[self.imp_var_name].unique())
            exporters = list(year_data[self.exp_var_name].unique())
            countries = set().union(importers, exporters)
            countries = list(countries)
            country_dict[year] = countries
        return country_dict
    
    
    def sector_list(self):
        return list(self.data_frame[self.sector_var_name].unique())

    def _extract_meta_data(self):
        '''
        Function to collect and assign meta data from data_summary object
        Returns: str
            Column names for importer, exporter, year, and sector fields
        '''
        return self.imp_var_name, self.exp_var_name, self.year_var_name, self.sector_var_name

    def add_note(self, note: str):
        '''
        Add a note to the 'notes' attribute
        Args:
            note: str
                A string to add to the list of notes
        Returns: None
        '''
        self.notes.append(note)

    def correlation(self, columns: List[str] = None, plot: bool = False):
        '''
        Return and plot the correlation matrix of the data.
        Args:
            columns: List[str] (optional)
                A list of column names to include in the matrix. Default is to include all columns.
            plot: bool
                If True, it plots a heatmap of the correlation matrix.

        Returns: Pandas.DataFrame
            A the correlation matrix
        '''
        if columns is None:
            corr_mat = self.data_frame.corr()
        else:
            corr_mat = self.data_frame[columns].corr()
        # Plot Correlation Matrix
        if plot:
            fig = plt.figure()
            ax = fig.add_subplot(111)
            cax = ax.matshow(corr_mat, interpolation='nearest', cmap = 'PiYG')
            fig.colorbar(cax)
            labels = list(corr_mat.columns.values)
            ax.set_xticklabels(labels, rotation='vertical')
            ax.set_yticklabels(labels)
            ax.set_xticks(np.arange(len(labels)))
            ax.set_yticks(np.arange(len(labels)))
            plt.show()
        return corr_mat

    def add_pair_var(self, var_list: List[str] = [], var_name: str = '', symmetric: bool = False):
        '''
        Create a new variable that is a concatenation of categorical variables for use as a fixed effect or cluster
        category, for example. Original values are separated by '_' in the concatenated value.
        Args:
            var_list: List[str]
                List of names of columns to concatinate.
            var_name: (Optional) str
                Column name to use for the new variable. By default, it uses the names in var_list separated by '_'.
            symmetric: bool
                If False, all variables are concatenated in the order of the var_list (e.g. ARG_USA
                and USA_ARG would both would be created). If True, ignores ordering of values across columns and
                concatenates alphabetically (e.g. all ARG and USA pairings would concatenate as ARG_USA)

        Returns: None, adds new column to EstimationData.dataframe.
        '''
        if len(var_name)==0:
            var_name = '_'.join(var_list)
        temp_data_frame = self.data_frame.loc[:, var_list].copy()
        for col in var_list:
            temp_data_frame[col] = temp_data_frame[col].astype(str)
        if symmetric:
            self.data_frame[var_name] = ['_'.join(x) for x in np.sort(temp_data_frame, axis=1)]
        else:
            self.data_frame[var_name] = temp_data_frame.apply('_'.join,axis=1)

        return None

    def __repr__(self):
         return "number of countries: {0} \n" \
                "number of exporters: {1} \n" \
                "number of importers: {2} \n" \
                "number of years: {3} \n" \
                "number of sectors: {4} \n" \
                "dimensions: {5}\n" \
             .format(self.number_of_countries,
                     self.number_of_exporters,
                     self.number_of_importers,
                     self.number_of_years,
                     self.number_of_sectors,
                     self.shape)










class _MetaData(object):
    '''
    A class that contains certain information to simplify its passing around different estimate routines.
    '''

    def __init__(self,
                 imp_var_name: str = 'importer',
                 exp_var_name: str = 'exporter',
                 year_var_name: str = 'year',
                 trade_var_name: str = None,
                 sector_var_name: str = None,
                 expend_var_name: str = None,
                 output_var_name: str = None,
                 ):
        self.imp_var_name = imp_var_name
        self.exp_var_name = exp_var_name
        self.year_var_name = year_var_name
        self.trade_var_name = trade_var_name
        self.sector_var_name = sector_var_name
        self.expend_var_name = expend_var_name
        self.output_var_name = output_var_name

    def __repr__(self):
        return "imp_var_name: {0} \n" \
               "exp_var_name: {1} \n" \
               "year_var_name: {2} \n" \
               "trade_var_name: {3} \n" \
               "sector_var_name: {4} \n" \
               "expend_var_name: {5}\n" \
               "output_var_name: {6} \n" \
            .format(self.imp_var_name,
                    self.exp_var_name,
                    self.year_var_name,
                    self.trade_var_name,
                    self.sector_var_name,
                    self.expend_var_name,
                    self.output_var_name)



__author__ = "Peter Herman"
__project__ = "gme.estimate"
__created__ = "05-16-2018"
__all__ = ['coefficient_kd_plot']

import pandas as pd
from .EstimationModel import EstimationModel
import matplotlib.pyplot as plt
from typing import List


def coefficient_kd_plot(estimation_model: EstimationModel,
                        variables: List[str],
                        path: str = None,
                        bandwidth: float = 0.5,
                        rename_variables: dict = None):
    """
    Produce kernel density plots of parameter estimates across different sectors in the results dictionary.

    Args:
        estimation_model: gme.EstimationModel
            An estimated EstimationModel with more than one sector.
        variables: List(str)
            A list of model covariates for which to plot kernel densities.
        path: (optional) str
            A path and file name at which to save the plot.
            Can end in the following file types for example: pdf, svg, and png.
        bandwidth: float
            Specify the bandwidth for the density plots. The default is 0.5.
        rename_variables: (optional)
            A dictionary of alternative variable names to use in the plot.
            For example {'original_name':'new_name}

    Returns: None
    """
    if estimation_model.results_dict is None:
        raise ValueError("results_dict does not exist. Must estimate model first.")

    results_dict = estimation_model.results_dict
    dict_key = results_dict.keys()
    coeff_df = pd.DataFrame(columns=variables)
    for var in variables:
        coefficients = []
        for key in dict_key:
            coefficients.append(results_dict[key].params[var])
        coeff_df[var] = pd.Series(coefficients)

    if rename_variables is not None:
        coeff_df.rename(columns=rename_variables, inplace=True)
    coeff_df.plot(kind='density', bw_method=bandwidth, subplots=True)

    if path is not None:
        plt.savefig(path)



__author__ = "Peter Herman"
__project__ = "gme.estimate"
__created__ = "04-30-2018"
__all__ = ['combine_sector_results']

import pandas as pd

def combine_sector_results(result_dict:dict = None,
                            write_path:str = None,
                            significance_stars: bool = False,
                            round_results: int = None,
                            latex_syntax: bool = False):
    '''
    Extract key result fields (coefficients, standard errors, and p-values) and combine them in a DataFrame. Has the option to write the data to a .csv file.
    Args:
        result_dict: Dict[statsmodels.genmod.generalized_linear_model.GLMResultsWrapper]
            A dictionary of GLM fit objects as returned by gme.EsimationModel.estimate()
        write_path: (optional) str
            A system location and file name in which to write a csv file containing the combined results.
        significance_stars: bool
            If true, combined results are output with significance stars. *** <0.01, **<0.05, and *<0.10. Default is False.
        round_results: (optional) int
            Rounds combined results to the desired decimal place.
        latex_syntax: bool
            If True, reports aspects of results, such as significance stars, using standard latex syntax.

    Returns: Pandas.DataFrame
        A DataFrame containing combined GLM results for all results in the supplied dictionary.

    Examples:
        # Using a gme.EstimationModel named sample_model...
        >>> sample_results = sample_model.estimate()
        >>> combine_sector_results(sample_results)
                                  all_coeff     all_pvalue              all_stderr
        log_distance              -0.739840  9.318804e-211  (0.023879411125052336)
        agree_pta                  0.334219   5.134355e-15   (0.04271952339258154)
        common_language            0.128770   1.076932e-03   (0.03938367074719932)
        contiguity                 0.255161   5.857612e-08   (0.04705076644539403)
        importer_year_fe_ARG2013  26.980367   0.000000e+00    (0.3612289201097519)

        >>> combine_sector_results(sample_results, path = 'c:\\Documents\\combined_results_saved.csv')
    '''


    keys = result_dict.keys()
    columns = []
    for column_name in keys:
        names = (str(column_name) + '_coeff',str(column_name) + '_stderr',str(column_name) + '_pvalue')
        sector = pd.DataFrame({
            names[0]: result_dict[column_name].params,
            names[1]: result_dict[column_name].bse,
            names[2]: result_dict[column_name].pvalues})

        one_percent = sector[names[2]] < 0.01
        five_percent = (sector[names[2]] < 0.05) & (sector[names[2]] >= 0.01)
        ten_percent = (sector[names[2]] < 0.1) & (sector[names[2]] >= 0.05)

        if round_results is not None:
            sector = sector.round(round_results)

        if significance_stars is True:
            if latex_syntax is True:
                sector.loc[one_percent, names[0]] = (sector[names[0]].astype(str) + '$^{***}$')
                sector.loc[five_percent, names[0]] = (sector[names[0]].astype(str) + '$^{**}$')
                sector.loc[ten_percent, names[0]] = (sector[names[0]].astype(str) + '$^{*}$')
            else:
                sector.loc[one_percent, names[0]] = (sector[names[0]].astype(str) + '***')
                sector.loc[five_percent, names[0]] = (sector[names[0]].astype(str) + '**')
                sector.loc[ten_percent, names[0]] = (sector[names[0]].astype(str) + '*')

        if latex_syntax is True:
            if round_results is not None:  # Rounding and converting to a string loses trailing zeros, this re-adds them
                sector[names[1]] = sector[names[1]].astype(str)
                lost_zero = (sector[names[1]].str.len() - sector[names[1]].str.index('.') - 1) < round_results
                sector.loc[lost_zero, names[1]] = (sector[names[1]] + '0')

            sector[names[1]] = ('(' + sector[names[1]].astype(str) + ')')

        singletons = {'nobs':result_dict[column_name].nobs,
                      'aic':result_dict[column_name].aic,
                      'bic': result_dict[column_name].bic,
                      'likelihood':result_dict[column_name].llf}

        for key in singletons.keys():
            new_row = pd.DataFrame({
            names[0]: singletons[key],
            names[1]: singletons[key],
            names[2]: singletons[key]}, index = [key])
            if round_results is not None:
                new_row = new_row.round(round_results)

            sector = pd.concat([sector,new_row], axis = 0)

        columns.append(sector)
    combined_sectors = pd.concat(columns, axis = 1)

    if write_path is not None:
        combined_sectors.to_csv(write_path)

    return combined_sectors


__author__ = "Saad Ahmad and Peter Herman"
__project__ = "gme.estimate"
__created__ = "05-24-2018"


import pandas as pd
import numpy as np
from typing import List, Tuple, Dict
from typing import Union
import statsmodels.api as sm
import time as time
import traceback

#-----------------------------------------------------------------------------------------#
# This file contains the underlying functions for the .estimate method in EstimationModel #
#-----------------------------------------------------------------------------------------#

# -----------
# Main function: Sequences the other routines
# -----------

def _estimate_ppml(data_frame,
                   meta_data,
                   specification,
                   fixed_effects: List[Union[str,List[str]]] = [],
                   drop_fixed_effect: Dict[Union[str, Tuple[str]],List[Union[str, Tuple[str]]]] = {},
                   cluster: bool=False,
                   cluster_on: str=None):
    '''
    Performs sector by sector GLM estimation with PPML diagnostics

    Args:
        data_frame: (Pandas.DataFrame) A dataframe containing data for estimation
        meta_data: (obj) a MetaData object from gme.EstimationData
        specification: (obj) a Specification object from gme.EstimationModel
        fixed_effects: (List[Union[str,List[str]]]) A list of variables to construct fixed effects based on.
            Can accept single string entries, which create fixed effects corresponding to that variable or lists of
            strings that create fixed effects corresponding to the interaction of the list items. For example,
            fixed_effects = ['importer',['exporter','year']] would create a set of importer fixed effects and a set of
            exporter-year fixed effects.
        drop_fixed_effect: (optional) Dict[Union[str,Tuple[str]],List[str]]
            A dictionary of fixed effect categories and names that are dropped from the fixed effects dataframe.
            drop_fixed_effect = {('importer','year'):[('ARG','2015'),('BEL','2013')]} would drop the
            fixed effects for importer-year: ARG-2015 and BEL-2013.
            The entry should be a subset of the list supplied for fixed_effects. 
    Returns: (Dict[GLM.fit], Pandas.DataFrame, Dict[DataFrame])
        1. Dictionary of statsmodels.GLM.fit objects with sectors as the keys.
        2. Dataframe with diagnostic information by sector
        3. Dictionary of estimation DataFrames + predicted trade values with sectors as the keys.
    '''

    post_diagnostics_data_frame_dict = {}
    results_dict = {}
    diagnostics_log = pd.DataFrame([])
    start_time = time.time()
    print('Estimation began at ' + time.strftime('%I:%M %p  on %b %d, %Y'))

    if not specification.sector_by_sector:
        data_frame = data_frame.reset_index(drop=True)
        fixed_effects_df = _generate_fixed_effects(data_frame, fixed_effects)
        fe_columns = list(fixed_effects_df.columns)
        estimating_data_frame = pd.concat([data_frame, fixed_effects_df], axis=1)
#       estimating_data_frame = pd.concat([data_frame[specification.lhs_var], data_frame[specification.rhs_var], fixed_effects_df], axis=1)
        model_fit, post_diagnostics_data_frame, diagnostics_log = _regress_ppml(estimating_data_frame, specification,fe_columns, drop_fixed_effect,cluster,cluster_on)
        
        results_dict['all'] = model_fit
                
        end_time = time.time()
        diagnostics_log.at['Completion Time'] = str(round((end_time - start_time)/60,2)) + ' minutes'
        post_diagnostics_data_frame_dict['all'] = post_diagnostics_data_frame


    else:
        sector_groups = data_frame.groupby(meta_data.sector_var_name)
        sector_list = _sectors(data_frame, meta_data)
        iteration_count = 1
        for sector in sector_list:
            sector_start_time = time.time()
            print('Sector ' + str(sector) + ' began at ' + time.strftime('%I:%M %p  on %b %d, %Y'))
            sector_data_frame = sector_groups.get_group(sector)
            sector_data_frame = sector_data_frame.reset_index(drop=True)

            # Create fixed effects
            fixed_effects_df = _generate_fixed_effects(sector_data_frame, fixed_effects)
            fe_columns = list(fixed_effects_df.columns)
            # Dataframe for estimations
            sector_data_frame = pd.concat([sector_data_frame, fixed_effects_df],axis=1)
            model_fit, post_diagnostics_data_frame, diagnostics_output = _regress_ppml(sector_data_frame, specification,fe_columns, drop_fixed_effect,cluster,cluster_on)

            # Timing reports
            sector_end_time = time.time()
            diagnostics_output.at['Sector Completion Time'] = (str(round((sector_end_time - sector_start_time) / 60, 2))
                                                               + ' minutes')
            if iteration_count > 1:
                average_time = ((time.time() - start_time) / 60) / iteration_count
                completion_time = (len(sector_list) - iteration_count) * average_time
                print("Average iteration time:  " + str(average_time) + "  minutes")
                print("Expected time to completion:  " + str(completion_time) + " minutes ("+ str(completion_time/60)
                      + " hours)\n")

            # Store results
            post_diagnostics_data_frame_dict[str(sector)] = post_diagnostics_data_frame
            results_dict[str(sector)] = model_fit
            diagnostics_log = pd.concat([diagnostics_log, diagnostics_output.rename(str(sector))], axis=1)
            iteration_count+=1

    print("Estimation completed at " + time.strftime('%I:%M %p  on %b %d, %Y'))
    return results_dict, diagnostics_log, post_diagnostics_data_frame_dict


# --------------
# Prep for Estimation Functions
# --------------
def _generate_fixed_effects(data_frame,
                            fixed_effects: List[Union[str, List[str]]] = []):
    '''
    Create fixed effects for single and interacted categorical variables.

    Args:
        data_frame: Pandas.DataFrame
            A DataFrame containing data for estimation.
        fixed_effects: List[Union[str,List[str]]]
            A list of variables to construct fixed effects based on.
            Can accept single string entries, which create fixed effects corresponding to that variable or lists of
            strings, which create fixed effects corresponding to the interaction of the list items. For example,
            fixed_effects = ['importer',['exporter','year']] would create a set of importer fixed effects and a set of
            exporter-year fixed effects.
    Returns: Pandas.DataFrame
        A DataFrame of fixed effects to be concatenated with the estimating DataFrame.
    '''

    fixed_effect_data_frame = pd.DataFrame([])
    # Get list for separate and combine fixed effect
    combined_fixed_effects = []
    separate_fixed_effects = []

    for item in fixed_effects:
        if type(item) is list:
            combined_fixed_effects.append(item)
        else:
            separate_fixed_effects.append(item)

    # Construct simple fixed effects
    for category in separate_fixed_effects:
        name = category + '_fe'
        temp_fe = pd.get_dummies(data_frame[category], prefix=name)
        fixed_effect_data_frame = pd.concat((fixed_effect_data_frame, temp_fe), axis=1)

    # Construct multiple fixed effects
    for item in combined_fixed_effects:
        if len(item) < 1:
            raise ValueError('A fixed_effects list element cannot be an empty list [].')

        if len(item) == 1:
            name = '_'.join(item) + '_fe'
            temp_fe = pd.get_dummies(data_frame[item[0]], prefix=name)
            fixed_effect_data_frame = pd.concat((fixed_effect_data_frame, temp_fe), axis=1)

        elif len(item) > 1:
            name = '_'.join(item) + '_fe'
            temp_data_frame = data_frame.loc[:, item]
            temp_data_frame.loc[:, name] = temp_data_frame.astype(str).sum(axis=1).copy()
            temp_fe = pd.get_dummies(temp_data_frame[name], prefix=name)
            fixed_effect_data_frame = pd.concat((fixed_effect_data_frame, temp_fe), axis=1)

    fixed_effect_data_frame = fixed_effect_data_frame.reset_index(drop=True)
    return fixed_effect_data_frame


def _drop_fe(data_frame,drop_dic):
    '''
    Drops user-provided fixed effect columns from a dataframe
    Arguments
    data_frame: Pandas.DataFrame
        A DataFrame containing data for estimation.
    drop_dic: Dict[Union[str,Tuple[str]],List[str]]
        A dictionary of fixed effect categories and names to be dropped from dataframe.
        drop_dic = {('importer','year'):[('ARG','2015'),('BEL','2013')]} would drop the
        fixed effect columns for importer-year: ARG-2015 and BEL-2013.
        The fixed effect categories should be a subset of the list supplied for fixed_effects. 
    Returns: (Pandas.DataFrame, List)
        1. A copy of the input data_frame with fixed effects columns removed
        2. List of fixed effect column names removed
    '''
        
    reduced_df=data_frame.copy()
    var_drop=[]
   
    #Check if values in dic are lists
    for key in drop_dic.keys():
        if isinstance(drop_dic[key],str):
                raise ValueError('Dropped Fixed Effect Names must be given as list or tuple')
    
    #Obtain the names to be dropped as a list of tuples
    for key in drop_dic.keys():
        for val in drop_dic[key]:
            if not isinstance(key,tuple):
                var_drop.append((key,val))
            else:   
                 var_drop.append(key+val)
   
    #Drop the columns from the dataframe that match the tuples
    column_drop=[]
    for var_tuple in var_drop:
        for col_name in reduced_df.columns:
            if all(drop_name in col_name for drop_name in var_tuple):
                reduced_df.drop(col_name, axis=1, inplace=True)
                column_drop.append(col_name)
    return reduced_df,column_drop

def _sectors(data_frame, meta_data):
    '''
    A function to extract a list of sectors from the estimating data_frame
    :param data_frame: (Pandas.DataFrame) A pandas data frmae to be used for estimation with a column defining
    sector or product IDs.
    :param meta_data: (obj) a MetaData object from gme.EstimationData
    :return: (list) a list of sorted sector IDs.
    '''
    sector_list = data_frame[meta_data.sector_var_name].unique()
    sector_list = np.ndarray.tolist(sector_list)
    sector_list.sort()
    return sector_list


# -------------
# PPML Regression and pre-diagnostics
# -------------


def _regress_ppml(data_frame, specification, fe_columns, drop_fixed_effect, cluster, cluster_on):
    '''
    Perform a GLM estimation with collinearity, insufficient variation, and overfit diagnostics and corrections.
    :param data_frame: (Pandas.DataFrame) A DataFrame for estimation
    :param specification: (obj) a Specification object from gme.EstimationModel
    :param specification: List of Fixed Effect columns in Dataframe
    :param drop_fixed_effect: (optional) A dictionary of FE categories and names to be dropped
    :return: (GLM.fit() obj, Pandas.DataFrame, Pandas.Series)
        1. The first returned object is a GLM.fit() results object containing estimates, p-values, etc.
        2. The second return object is the dataframe used for estimation that has problematic columns removed.
        3. A column containing diagnostic information from the different checks and corrections undertaken.
    '''
    # Check for zero trade fixed effects
    data_frame_copy = data_frame.copy()
    adjusted_data_frame, problem_variable_list, rhs_cols = _new_trade_contingent_collinearity_check(data_frame=data_frame_copy,
                                                                                      specification=specification, fe_columns=fe_columns)

    # Check for perfect collinearity and drop any user-specified FE
    rhs = adjusted_data_frame[rhs_cols] 
    adj_rhs,user_fe=_drop_fe(rhs,drop_fixed_effect)
    non_collinear_rhs, collinear_fe = _collinearity_check(adj_rhs)

    total_fe_drop=user_fe+collinear_fe
    for col in total_fe_drop:
        adjusted_data_frame.drop(col, 1, inplace=True)
#
#    if len(collinear_column_list) == 0:
#        collinearity_indicator = 'No'
#    else:
#        collinearity_indicator = 'Yes'
    #collinearity_column = pd.Series({'Collinearities': collinearity_indicator})
    excluded_column_list = problem_variable_list + total_fe_drop
    exclusion_column = pd.Series({'Number of Regressors Dropped': len(excluded_column_list)})
    print('Omitted Regressors: ' + str(excluded_column_list))
   
    # GLM Estimation
    if cluster is False:
        try:
            estimates = sm.GLM(endog=adjusted_data_frame[specification.lhs_var],
                           exog=non_collinear_rhs,
                           family=sm.families.Poisson()
                           ).fit(cov_type=specification.std_errors,
                                 maxiter=specification.iteration_limit)
            adjusted_data_frame['predicted_trade'] = estimates.mu

        except:
            traceback.print_exc()
            estimates = 'Estimation could not complete.  GLM process raised an error.'
    else:
        try:

            model = sm.GEE(endog=adjusted_data_frame[specification.lhs_var],
                        exog=non_collinear_rhs,groups=adjusted_data_frame[cluster_on], 
                        family=sm.families.Poisson())
            
            estimates=model.fit()
            adjusted_data_frame['predicted_trade'] = model.predict(estimates.params)

        except:
            traceback.print_exc()
            estimates = 'Estimation could not complete.  GLM process raised an error.'
        
        
    # Checks for overfit (only valid when keep=False)
    try:
        fit_check_outcome = _overfit_check(data_frame=adjusted_data_frame,
                                           specification=specification,
                                           predicted_trade_column='predicted_trade')
        overfit_column = pd.Series({'Overfit Warning': fit_check_outcome})
    except:
        # CHECK THAT ESTIMATES DOES NOT GET USED LATER
        # Add something to return for diagnostics. I believe it needs to be a dataframe that can be combined
        overfit_column = pd.Series({'Overfit Warning': 'Estimation could not complete'})


    # Collect diagnostics
    diagnostics = overfit_column.append(exclusion_column)
    #diagnostics = diagnostics.append(exclusion_column)
    diagnostics.at['Regressors with Zero Trade'] =  problem_variable_list
    diagnostics.at['Regressors from User'] = user_fe
    diagnostics.at['Regressors Perfectly Collinear'] = collinear_fe
    
    return estimates, adjusted_data_frame, diagnostics


def _trade_contingent_collinearity_check(data_frame, specification):
    '''
    PPML diagnostic for columns that are collinear when trade is greater than zero, as in Santos and Silva (2011)
    Arguments
    :param data_frame: (Pandas.DataFrame) A DataFrame for estimation
    :param specification: (obj) a Specification object from gme.EstimationModel
    :return: (Pandas.DataFrame, list)
        1. A copy of the input data_frame with columns collinear when trade is greater than zero and associated
        observations removed
        2. List containing the names of the columns that were collinear when trade is greater than zero.
    '''

    # Main dataframe for manipulation
    data_frame_copy = data_frame.copy()

    # Identify problematic variables due to perfect collinearity when y>0
    nonzero_data_frame = data_frame_copy.loc[data_frame_copy[specification.lhs_var] > 0,:]
    lhs = [specification.lhs_var]
    rhs_columns = list(nonzero_data_frame.columns)
    rhs_columns.remove(specification.lhs_var)
    rhs = nonzero_data_frame[rhs_columns]
    noncollinear_columns, excluded_columns_list = _collinearity_check(rhs)
    rhs_columns = list(noncollinear_columns.columns)

    # Check if problematic and delete associated observations
    data_frame_copy['mask'] = 1
    problem_variable_list = []
    for col in excluded_columns_list:
        # mean_value = data_frame.loc[(data_frame_copy[specification.lhs_var] > 0),col].mean()
        mean_value = data_frame_copy[data_frame_copy[specification.lhs_var] > 0][col].mean()
        max_value = data_frame_copy[data_frame_copy[specification.lhs_var] == 0][col].max()
        min_value = data_frame_copy[data_frame_copy[specification.lhs_var] == 0][col].min()
        if min_value < mean_value and mean_value < max_value:
            rhs_columns.append(col)
        else:
            problem_variable_list.append(col)
            if data_frame_copy[col].nunique() == 2:
                data_frame_copy.loc[data_frame_copy[col] == 1, 'mask'] = 0

    # Return final data_frame with removed columns and observations
    data_frame_copy = data_frame_copy.loc[data_frame_copy['mask'] != 0,:]
    all = lhs + rhs_columns
    data_frame_copy = data_frame_copy[all]
    return data_frame_copy, problem_variable_list


def _collinearity_check(data_frame, tolerance_level=1e-05):
    '''
    Identifies and drops perfectly collinear columns
    :param data_frame: (Pandas.DataFrame) A DataFrame for estimation
    :param tolerance_level: (float) Tolerance parameter for identifying zero values (default=1e-05)
    :return: (Pandas.DataFrame) Original DataFrame with collinear columns removed
    '''

    data_array = data_frame.values
    q_factor, r_factor = np.linalg.qr(data_array)
    r_diagonal = np.abs(r_factor.diagonal())
    r_range = np.arange(r_diagonal.size)

    # Get list of collinear columns
    collinear_columns = np.where(r_diagonal < tolerance_level)[0]
    collinear_data_frame = data_frame.iloc[:, collinear_columns]
    collinear_column_list = list(collinear_data_frame.columns.values)

    # Get df with independent columns
    collinear_locations = list(collinear_columns)
    independent_columns = list(set(r_range).symmetric_difference(set(collinear_locations)))
    return data_frame.iloc[:, independent_columns], collinear_column_list


# ----------------
# Post-estimation Diagnostics
# ----------------

def _overfit_check(data_frame, specification: str = 'trade',
                   predicted_trade_column: str = 'predicted_trade'):
    '''
    Checks if predictions from GLM estimation are perfectly fitted arguments
    :param data_frame: (Pandas.DataFrame)
    :param specification:
    :param predicted_trade_column:
    :return:
    '''
    '''
    Checks if predictions from GLM estiamtion are perfectly fitted
    Arguments
    ---------
    first: DataFrame object
    second: String for variable name containing trade values (default='trade')
    third: String for variable name containing predicted trade values (defalt='ptrade')

    Returns
    -------
    String indicting if the predictions are perfectly fitted
    '''
    fit_check = 'No'
    non_zero = data_frame.loc[data_frame[specification.lhs_var] > 0,specification.lhs_var]
    low = non_zero.min() * 1e-6
    fit_zero_observations = data_frame.loc[data_frame[specification.lhs_var] == 0, predicted_trade_column]
    if fit_zero_observations.min() < low:
        fit_check = 'Yes'
    return fit_check

def _new_trade_contingent_collinearity_check(data_frame, specification, fe_columns):
    '''
    PPML diagnostic for columns that are collinear when trade is greater than zero, as in Santos and Silva (2011)
    Arguments
    :param data_frame: (Pandas.DataFrame) A DataFrame for estimation
    :param specification: (obj) a Specification object from gme.EstimationModel
    :fe_columns: List of fe_variables to be used in the estimation
    :return: (Pandas.DataFrame, list)
        1. A copy of the input data_frame with columns collinear when trade is greater than zero and associated
        observations removed
        2. List containing the names of the columns that were collinear when trade is greater than zero.
        3. List of remaining fixed effect columns
    '''

    # Main dataframe for manipulation
    data_frame_copy = data_frame.copy()

    # Identify problematic variables due to perfect collinearity when y>0
    nonzero_data_frame = data_frame_copy.loc[data_frame_copy[specification.lhs_var] > 0]

    orig_rhs_columns = specification.rhs_var + fe_columns
    rhs = nonzero_data_frame[orig_rhs_columns]
    noncollinear_columns, excluded_columns_list = _collinearity_check(rhs)
    new_rhs_columns = list(noncollinear_columns.columns)

    # Check if problematic and delete associated observations
    data_frame_copy['mask'] = 1
    problem_variable_list = []
    for col in excluded_columns_list:
        mean_value = data_frame_copy[data_frame_copy[specification.lhs_var] > 0][col].mean()
        max_value = data_frame_copy[data_frame_copy[specification.lhs_var] == 0][col].max()
        min_value = data_frame_copy[data_frame_copy[specification.lhs_var] == 0][col].min()
        if min_value < mean_value and mean_value < max_value:
            new_rhs_columns.append(col)
        else:
            problem_variable_list.append(col)
            if data_frame_copy[col].nunique() == 2:
                data_frame_copy.loc[data_frame_copy[col] == 1, 'mask'] = 0

    # Return final data_frame with removed columns and observations
    #data_frame_copy = data_frame_copy[data_frame_copy['mask'] != 0] #To account for FEs with non-zero trade
    drop_obs= data_frame_copy[(data_frame_copy[specification.lhs_var] == 0) & (data_frame_copy['mask']== 0)].index
    data_frame_copy.drop(drop_obs,inplace=True)
    
    drop_columns=problem_variable_list + ['mask']
    for col in drop_columns:
        data_frame_copy.drop(col, axis=1, inplace=True)

    return data_frame_copy, problem_variable_list, new_rhs_columns









__author__ = "Peter Herman"
__project__ = "gme.estimate"
__created__ = "05-07-2018"
__all__ = ['format_regression_table']
from typing import List
import pandas as pd


def format_regression_table(results_dict: dict = None,
                            variable_list: List[str] = [],
                            format: str = 'txt',
                            se_below: bool = True,
                            significance_levels: List[float] = [0.01, 0.05, 0.10],
                            round_values: int = 3,
                            omit_fe_prefix: List[str] = [],
                            table_columns: list = [],
                            path: str = None,
                            include_index: bool = False,
                            latex_syntax: bool = False,
                            r_squared: bool = False,
                            note: str = None):
    '''
    Format estimation results into a standard table format with options for significance stars, LaTeX syntax, standard
    error positioning, rounding, fixed effect ommission, and others options.

    Args:
        results_dict: Dict[statsmodels.genmod.generalized_linear_model.GLMResultsWrapper]
            A dictionary of GLM fit objects from statsmodels
        variable_list: (optional) List[str]
            A list of variables to include in the results table. If none are provided, all variables are included. The
            default is an empty list, which results in the inclusion of all estimated variables.
        format: str
            Determines the file formatting of text. Accepts 'tex' for LaTeX, 'txt' for plain text, or 'csv' for a
            csv table. Default is 'txt'.
        se_below: bool
            If True, standard errors are presented below estimates. If False, they are presented in a column to the
            right. The default is True.
        significance_levels: List[float]
            A list specifying the three percentages, from lowest to highest, on which to base significance stars. The
            default value is [0.01, 0.05, 0.10].
        round_values: int
            The number of decimal points to include in the reported figures. The default is 3.
        omit_fe_prefix: (optional) List[str]
            A list of strings such that any variable starting with that string are omitted from the created table. The
            value is an empty list that omits no variables.
        table_columns: (optional) List[str]
            A list of keys from the results_dict to be included in the created table. The default is an empty list, which
            results in all values being created
        path: (optional) str
            A system path and file name to write the created table to.  File extensions of .txt (format = 'txt'),
            .tex or .txt (format = 'tex'), or .csv (format = 'csv') are recommended.
        include_index: bool
            If true, the outputed .csv file will contain row numbers. Default is False.
        latex_syntax: bool
            If true, the table will include LaTeX syntax, regardless of the chosen format. Default is False
        variable_order: (optional) List[str]
            If supplied, provides an specific ordering in which to list the variables in the table.
        r_squared:  bool
            If True, it includes R^2 values in the table. This is primarily useful if OLS regression results are
            supplied. Default is False.
        note: (optional) str
            Adds the supplied string as a note at the bottom of the table.

    Returns: Pandas.DataFrame
        A DataFrame containing the formatted results table with specified syntax.

    Examples:
        Create a .csv file.
        >>> sample_estimation_model.format_regression_table(format = 'csv',
                                                            path = "c:\folder\saved_results.csv")

        Create a LaTeX .tex table without fixed effects (with prefix 'imp_fe_' and 'exp_fe_')
        >>> sample_estimation_model.format_regression_table(format = 'tex',
                                                            path = "c:\folder\saved_results.tex",
                                                            omit_fe_prefix = ['imp_fe_' , 'exp_fe_'])
    '''

    if results_dict is None:
        raise ValueError('Must input a dictionary of regression (GLM.fit) objects')

    if len(table_columns) == 0:
        table_columns = list(results_dict.keys())
    else:
        for column in table_columns:
            if column not in list(results_dict.keys()):
                raise ValueError('Specified column {0} in table_columns is not a key in results_dict.'.format(column))

    formatted_dict = {}
    for key in table_columns:

        results = results_dict[key]
        if se_below is True:
            compiled_results = pd.DataFrame(columns=['Variable', str(key)])
        if se_below is False:
            compiled_results = pd.DataFrame(columns=['Variable', str(key), (str(key) + ' SE')])

        if len(variable_list) == 0:
            variable_list_current = results.params.index
        else:
            variable_list_current = pd.Series(variable_list)

        if len(omit_fe_prefix) > 0:
            for prefix in omit_fe_prefix:
                variable_list_current = variable_list_current[~variable_list_current.str.startswith(prefix)]

        for variable in variable_list_current:

            # Add significance stars to coefficient
            beta = str(round(results.params[variable], round_values))
            while (len(beta) - beta.index('.') - 1) < round_values:  # pad trailing zeros if dropped
                beta = beta + '0'
            if format == 'tex' or latex_syntax is True:
                if results.pvalues[variable] < significance_levels[0]:
                    formatted_coeff = beta + '$^{***}$'
                if (results.pvalues[variable] < significance_levels[1]) & \
                        (results.pvalues[variable] >= significance_levels[0]):
                    formatted_coeff = beta + '$^{**}$'
                if (results.pvalues[variable] < significance_levels[2]) & \
                        (results.pvalues[variable] >= significance_levels[1]):
                    formatted_coeff = beta + '$^{*}$'
                if results.pvalues[variable] >= significance_levels[2]:
                    formatted_coeff = beta
            else:
                if results.pvalues[variable] < significance_levels[0]:
                    formatted_coeff = beta + '***'
                if (results.pvalues[variable] < significance_levels[1]) & \
                        (results.pvalues[variable] >= significance_levels[0]):
                    formatted_coeff = beta + '**'
                if (results.pvalues[variable] < significance_levels[2]) & \
                        (results.pvalues[variable] >= significance_levels[1]):
                    formatted_coeff = beta + '*'
                if results.pvalues[variable] >= significance_levels[2]:
                    formatted_coeff = beta

            # Format standard error
            std_err = str(round(results.bse[variable], round_values))
            while (len(std_err) - std_err.index('.') - 1) < round_values:  # pad trailing zeros if dropped
                std_err = std_err + '0'
            formatted_se = '(' + std_err + ')'

            if se_below is False:
                row = pd.DataFrame({'Variable': variable,
                                    str(key): formatted_coeff,
                                    (str(key) + ' SE'): formatted_se},
                                   index=[('a_' + variable)])
                compiled_results = pd.concat([compiled_results, row], axis=0)

            if se_below is True:
                row = pd.DataFrame({'Variable': [variable, ' '], str(key): [formatted_coeff, formatted_se]},
                                   index=['a_' + variable, 'a_' + str(variable) + '_se'])
                compiled_results = pd.concat([compiled_results, row], axis=0)

        # GEE models used for cluster option do not report some stats so must catch it and replace with something else
        try:
            num_obs = str(int(results.nobs))
            aic = str(round(results.aic, round_values))
            bic = str(round(results.bic, round_values))
            llf = str(round(results.llf, round_values))
        except:
            num_obs = 'Not reported for model type'
            aic = num_obs
            bic = num_obs
            llf = num_obs
        if se_below is False:
            row = pd.DataFrame({'Variable': ['Obs.', 'AIC', 'BIC', 'Likelihood'],
                                str(key): [num_obs,
                                           aic,
                                           bic,
                                           llf],
                                (str(key) + ' SE'): ['', '', '', '']},
                               index=['b_nobs', 'b_aic', 'b_bic', 'b_llf'])
            if r_squared is True:
                row_r_squared = pd.DataFrame({'Variable': ['R^2'],
                                              str(key): [str(round(results.rsquared, 4))],
                                              (str(key) + ' SE'): ['']},
                                             index=['b_R2'])
                row = pd.concat([row, row_r_squared], axis=0)

        if se_below is True:
            row = pd.DataFrame({'Variable': ['Obs.', 'AIC', 'BIC', 'Likelihood'],
                                str(key): [num_obs,
                                           aic,
                                           bic,
                                           llf]},
                               index=['b_nobs', 'b_aic', 'b_bic', 'b_llf'])
            if r_squared is True:
                row_r_squared = pd.DataFrame({'Variable': ['R^2'],
                                              str(key): [str(round(results.rsquared, 4))]},
                                             index=['b_R2'])
                row = pd.concat([row, row_r_squared], axis=0)

        compiled_results = pd.concat([compiled_results, row], axis=0)

        formatted_dict[key] = compiled_results

    results_table = formatted_dict[table_columns[0]]
    remaining_columns = table_columns[1:]
    for key in remaining_columns:
        results_table = results_table.merge(right=formatted_dict[key], how='outer', left_index=True, right_index=True)
        missing_var_name = results_table['Variable_x'].isnull()
        results_table.loc[missing_var_name, 'Variable_x'] = results_table['Variable_y']
        results_table.drop(['Variable_y'], axis=1, inplace=True)
        results_table.rename(columns={'Variable_x': 'Variable'}, inplace=True)

    results_table.sort_index(inplace=True)

    # Idea: custom sort order for variables

    results_table[results_table.isnull()] = ''

    # Reorder 'Variable' column first
    column_names = results_table.columns.tolist()
    column_names.insert(0, column_names.pop(column_names.index('Variable')))
    results_table = results_table[column_names]
    if path is not None:
        if format == 'tex':
            column_spec = '{' + results_table.shape[1] * 'l' + '}'
            begin_tab = '\\begin{tabular}' + column_spec
            end_tab = '\\end{tabular}'
            results_table.to_csv(path, index=include_index, sep='&', line_terminator='\\\\\n')
            with open(path, 'r+') as file:
                content = file.read()
                content = content.replace('_', '\\_')
                file.seek(0, 0)
                file.write(begin_tab.rstrip('\r\n') + '\n' + content)
            with open(path, 'a+') as file:
                file.write(end_tab)

        if format == 'txt':
            with open(path, 'w') as file:
                file.write(results_table.to_string(index=include_index, justify='left'))

        if format == 'csv':
            results_table.to_csv(path, index=include_index, sep=',', line_terminator='\n')

        # Add note
        footnote = '*** p < {}, ** p < {}, * p < {}. '.format(significance_levels[0], significance_levels[1],
                                                              significance_levels[2])
        with open(path, 'a+') as file:
            if format == 'csv':
                file.write('\n"{}"'.format(footnote))
                if note is not None:
                    file.write('"{}"'.format(note))
            else:
                file.write("\n{}".format(footnote))
                if note is not None:
                    file.write(note)

    return results_table



__Author__ = "Peter Herman"
__Project__ = "Construct Data"
__Created__ = "01/11/2018"
__all__ = ['make_data_square']

import numpy as np
import pandas as pd
from warnings import warn


def make_data_square(trade_dataframe=None,
                     imp_var_name: str = "importer",
                     exp_var_name: str = "exporter",
                     multiple_sectors: bool = False,
                     sector_var_name: str = None,
                     year_var_name: str = "year",
                     drop_intratrade: bool = True,
                     year_by_year: bool = True):
    """
    Expand countries and sectors by adding zero-valued trade flows to make the panel square.
    Args:
        trade_dataframe: pd.DataFrame
            A DataFrame containing trade values and observation identifiers (importer, exporter, year, and (optionally)
            sector).
        imp_var_name: str
            The name of the column containing importer identifiers.
        exp_var_name: str
            The name of the column containing exporter identifiers.
        multiple_sectors: bool
            If true, the data is squared for each sector as well. Default is False.
        sector_var_name: (optional) str
            The name of the column containing sector identifiers. Must be supplied if squaring for each section (i.e.
            if multiple_sectors = True).
        year_var_name: (optional) str
            The name of the column containing year identifiers.
        drop_intratrade: bool
            If True, the returned data will not contain intranational observations in which the importer and exporter
            are the same country. Default is True.
        year_by_year: bool
            If true, the data is squared on a year by year basis, which allows the dimensions to differ each year. For
            example, it lets the set of countries or sectors around which the data is squared to differ each year.
            Doing so reduces the likelihood of including countries or sectors with zero trade activity in certain years.

    Returns: pd.DataFrame
        A square dataset.

    """
    if (multiple_sectors and sector_var_name is None):
        raise ValueError("A column name for sectors is required if multiple_sectors == True")

    if not year_by_year:
        complete_square_data = _not_year_by_year(trade_dataframe, imp_var_name, exp_var_name,
                                                 multiple_sectors, sector_var_name, year_var_name, drop_intratrade)

    else:
        complete_square_data = _year_by_year(trade_dataframe,
                                             imp_var_name,
                                             exp_var_name,
                                             multiple_sectors,
                                             sector_var_name,
                                             year_var_name,
                                             drop_intratrade)

    return complete_square_data


def _not_year_by_year(trade_dataframe,
                      importer_var_name: str,
                      exporter_var_name: str,
                      multiple_sectors: bool,
                      sector_var_name: str,
                      year_var_name: str,
                      drop_intratrade: bool):
    '''
    Creates a square dataset based on observation present in any year.
    :param trade_dataframe: (Pandas DataFrame) A Pandas DataFrame containing bilateral trade data
    :param importer_var_name: (str) The name of the column containing importer IDs.
    :param exporter_var_name: (str) The name of the column containing exporter IDs.
    :param multiple_sectors: (bool) True if the data contains multiple sectors. Default is False.
    :param sector_var_name: (str) The name of the column containing sector IDs, if applicable.
    :param year_var_name: (str) The name of the column containing year IDs, if applicable
    :param drop_intratrade: (bool) True if you would like intra-country observations dropped (i.e. importer == exporter). Default is True
    :return: a square (Pandas DataFrame) object
    '''
    countries = _country_list(trade_dataframe, importer_var_name, exporter_var_name)
    countries['key'] = 0  # this provides a series to merge on.

    square_data = countries.merge(countries, how='outer', on='key')
    square_data.rename(index=str, columns={"0_x": importer_var_name, "0_y": exporter_var_name}, inplace=True)

    merging_dimensions = [importer_var_name, exporter_var_name]
    expected_length = len(countries)
    expected_width = 3

    if multiple_sectors:
        square_data, merging_dimensions, expected_length, expected_width = _square_multiple_sectors(trade_dataframe,
                                                                                                    sector_var_name,
                                                                                                    merging_dimensions,
                                                                                                    expected_length,
                                                                                                    expected_width)

    square_data, merging_dimensions, expected_length, expected_width = _square_years(trade_dataframe, square_data,
                                                                                     year_var_name, merging_dimensions,
                                                                                     expected_length, expected_width)

    if drop_intratrade:
        square_data = _drop_intratrade(square_data, importer_var_name, exporter_var_name)
        expected_length = expected_length * (len(countries) - 1)
    else:
        expected_length = expected_length * len(countries)

    square_data.drop('key', axis=1, inplace=True)
    square_data = square_data.merge(trade_dataframe, how='left', on=merging_dimensions)
    square_data.fillna(0, inplace=True)

    if square_data.shape != (expected_length, expected_width):
        warn(
            "dimensions of the constructed square data ({0}, {1}) do not match the expected dimensions ({2}, {3})".format(
                square_data.shape[0], square_data.shape[1], expected_length, expected_width))

    print("Number of Missing Values in Square Trade Data")
    print(square_data.isnull().sum())

    complete_square_data = square_data

    return complete_square_data


def _year_by_year(trade_dataframe,
                  importer_var_name: str,
                  exporter_var_name: str,
                  multiple_sectors: bool,
                  sector_var_name: str,
                  year_var_name: str,
                  drop_intratrade: bool):
    '''
    Creates a square dataset such that each year is squared based on the observations present only in that year.
    :param trade_dataframe: (Pandas DataFrame) A Pandas DataFrame containing bilateral trade data
    :param importer_var_name: (str) The name of the column containing importer IDs.
    :param exporter_var_name: (str) The name of the column containing exporter IDs.
    :param multiple_sectors: (bool) True if the data contains multiple sectors. Default is False.
    :param sector_var_name: (str) The name of the column containing sector IDs, if applicable.
    :param year_var_name: (str) The name of the column containing year IDs, if applicable
    :param drop_intratrade: (bool) True if you would like intra-country observations dropped (i.e. importer == exporter). Default is True
    :return: a square (Pandas DataFrame) object
    '''
    years = _year_list(trade_dataframe, year_var_name)
    grouped_data = trade_dataframe.groupby(year_var_name)
    iteration_counter = 1

    for year_group in years:
        print(year_group)

        year_data = grouped_data.get_group(year_group)
        countries = _country_list(year_data, importer_var_name, exporter_var_name)

        print("Number of Countries in " + str(year_group) + ': ' + str(len(countries)))
        countries['key'] = 0  # this provides a series to merge on.

        square_data = countries.merge(countries, how='outer', on='key')
        square_data.rename(index=str, columns={"0_x": importer_var_name, "0_y": exporter_var_name}, inplace=True)

        merging_dimensions = [importer_var_name, exporter_var_name, year_var_name]
        expected_length = len(countries)
        expected_width = 3

        if multiple_sectors:
            square_data, merging_dimensions, expected_length, expected_width = _square_multiple_sectors(year_data,
                                                                                                        square_data,
                                                                                                        sector_var_name,
                                                                                                        merging_dimensions,
                                                                                                        expected_length,
                                                                                                        expected_width)

        square_data[year_var_name] = int(year_group)
        expected_width += 1

        if drop_intratrade:
            square_data = _drop_intratrade(square_data, importer_var_name, exporter_var_name)
            expected_length = expected_length * (len(countries) - 1)
        else:
            expected_length = expected_length * len(countries)

        square_data.drop('key', axis=1, inplace=True)

        square_data = square_data.merge(year_data, how='left', on=merging_dimensions)
        square_data.fillna(0, inplace=True)

        if square_data.shape != (expected_length, expected_width):
            warn(
                "dimensions of the constructed square data ({0}, {1}) do not match the expected dimensions ({2}, {3})".format(
                    square_data.shape[0], square_data.shape[1], expected_length, expected_width))

        print("Number of Missing Values in Square Trade Data")
        print(square_data.isnull().sum())

        if iteration_counter == 1:
            complete_square_data = square_data
        else:
            complete_square_data = complete_square_data.append(square_data)
        iteration_counter += 1

    return complete_square_data


def _year_list(dataframe, year_var_name):
    '''
    Create a list of years in the data
    :param dataframe:(Pandas DataFrame) dataframe to extract years from
    :param year_var_name: (str) name of column containing year IDs
    :return: a list of years in the data
    '''
    years = pd.DataFrame(dataframe[year_var_name].unique())
    return years[0].tolist()


def _country_list(dataframe, importer_var_name, exporter_var_name):
    '''
    Create a list of importers and exporters in the data
    :param dataframe: (Pandas DataFrame) dataframe to extract countries from
    :param importer_var_name: (str) name of column containing importer IDs
    :param exporter_var_name: (str) name of column containing exporter IDs
    :return: a list of all countries in the data
    '''
    countries = np.concatenate((dataframe[importer_var_name].unique(), dataframe[exporter_var_name].unique()),
                               axis=0)
    return pd.DataFrame(pd.DataFrame(countries)[0].unique())


def _square_multiple_sectors(original_dataframe, square_dataframe, sector_var_name, merging_dimensions, expected_length,
                             expected_width):
    '''
    Squares a dataframe along the 'sectors' dimension.
    :param original_dataframe: (Pandas DataFrame) Dataframe to extract a sector list from
    :param square_dataframe: (Pandas DataFrame) Square dataframe to expand to sectors
    :param sector_var_name: (str) name of column containing sector IDs
    :param merging_dimensions: (List) A list containing the column names to merge observed data back onto
    :param expected_length: (int) expected number of rows in square_dataframe
    :param expected_width: (int) expected number of columns in square_dataframe
    :return: square_dataframe - (Pandas Dataframe)  updated version of input square_dataframe with sectors squared too.
             merging_dimensions - (List) updated...
             expected_length - (int) updated...
             expected_width - *int) updated...
    '''
    sectors = pd.DataFrame(original_dataframe[sector_var_name].unique())
    sectors['key'] = 0
    square_dataframe = square_dataframe.merge(sectors, how='outer', on='key')
    square_dataframe.rename(index=str, columns={0: sector_var_name}, inplace=True)
    merging_dimensions.append(sector_var_name)
    expected_length = expected_length * len(sectors)
    expected_width += 1
    return square_dataframe, merging_dimensions, expected_length, expected_width


def _square_years(original_dataframe, square_dataframe, year_var_name, merging_dimensions, expected_length,
                  expected_width):
    '''
    Sqaures a dataframe along the 'years' dimension.
    :param original_dataframe: (Pandas DataFrame) Dataframe to extract a year list from
    :param square_dataframe: (Pandas DataFrame) Square dataframe to expand to years
    :param year_var_name: (str) name of column containing year IDs
    :param merging_dimensions: (List) A list containing the column names to merge observed data back onto
    :param expected_length: (int) expected number of rows in square_dataframe
    :param expected_width: (int) expected number of columns in square_dataframe
    :return: square_dataframe - (Pandas Dataframe)  updated version of input square_dataframe with years squared too.
             merging_dimensions - (List) updated...
             expected_length - (int) updated...
             expected_width - *int) updated...
    '''
    years = pd.DataFrame(original_dataframe[year_var_name].unique())
    years['key'] = 0
    square_dataframe = square_dataframe.merge(years, how='outer', on='key')
    square_dataframe.rename(index=str, columns={0: year_var_name}, inplace=True)
    merging_dimensions.append(year_var_name)
    expected_length = expected_length * len(years)
    expected_width += 1
    return square_dataframe, merging_dimensions, expected_length, expected_width


def _drop_intratrade(dataframe,
                     importer_var_name,
                     exporter_var_name):
    '''
    Drops intratrade observations (i.e. importer == exporter)
    :param dataframe: (Pandas DataFrame)
    :param importer_var_name: (str) The name of the column containing importer IDs.
    :param exporter_var_name: (str) The name of the column containing exporter IDs.
    :return: (Pandas DataFrame)
    '''
    dataframe.drop(
        dataframe.loc[dataframe[importer_var_name] == dataframe[exporter_var_name]].index,
        inplace=True)
    return dataframe



__Author__ = "Peter Herman"
__Project__ = "Gravity Code"
__Created__ = "03/12/2018"
__all__ = ['DiagnosticsLog']

import datetime

class DiagnosticsLog(object):

    def __init__(self,
                 spec_name:str = 'default_name',
                 ):
        self.creation_date = datetime.datetime.now().strftime("%y-%m-%d, %H:%M")
        self.spec_name = spec_name


from .construct_data.DataLog import *
from .construct_data.EstimationData import *
from .construct_data.make_data_square import *
from .estimate.combine_sector_results import *
from .estimate.DiagnosticsLog import *
from .estimate.EstimationModel import *
from .estimate.format_regression_table import *
from .estimate.save_and_load import *
from .estimate.SlimResults import *
from .estimate.Specification import *
from .estimate.visualize_results import *



__Author__ = "Peter Herman"
__Project__ = "Gravity Code"
__Created__ = "03/12/2018"
__all__ = ['DataLog']

import datetime

class DataLog(object):

    def __init__(self,
                 parent_logs = []):
        self.parent_logs = parent_logs
        self.creation_date = datetime.datetime.now().strftime("%y-%m-%d, %H:%M")

        # --------------------
        # Part of _estimate()
        # --------------------

        # Part of _slice_data_for_estimation
        self.intra_country_trade_dropped = 'not affected'
        self.importers_dropped = 'not_affected'
        self.exporters_dropped = 'not affected'
        self.importers_kept = 'not affected'
        self.exporters_kept = 'not affected'
        self.years_dropped = 'not affected'
        self.years_kept = 'not affected'
        self.missing_dropped = 'not affected'
